module Relation (module Relation) where

import ST

infixl 3 $$ , .<.


------

class (Eq a) => Spanned a where
     s_span :: [a]



---------------------------------------------------------------------------------------------------
------ Cpo
---------------------------------------------------------------------------------------------------

class Cpo c  where
   top   :: c
   bot   :: c
--
   up  :: c -> [c]
--  define "up"  or ".<." + "range"
--
   (.<.) :: (Eq c) => c -> c -> Bool
   x .<. y = y `elem` (up x)                    -- default
--
   range :: (Eq c) => [c]
   range = cover up bot                         -- default



--------------------------------------------------------------------------------------------------


instance (Spanned a, Spanned b) => Spanned (a,b) where
     s_span = [(a,b) | a <- s_span , b <- s_span ]

instance Spanned a => Spanned [a] where
     s_span = power s_span

instance Spanned a => Cpo [a]  where
    top     = s_span
    bot     = []
    range   = power s_span
    u .<. v = all (`elem` v) u
    up u     = map suc s_span where
	          suc a = if a `elem` u then u else (a:u)

----------------------------------------------------------------------------------------------------
-- Relations
----------------------------------------------------------------------------------------------------

data Rel a b = Rel { ($$) :: a -> [b] }

listToRel :: (Eq a) => [(a,b)] -> Rel a b
listToRel l = Rel (\x -> [b | (a,b) <- l , x == a])
--
relToList :: (Spanned a) => Rel a b -> [(a,b)]
relToList u = [(a,b) | a <- s_span , b <- u $$ a ]

------- Instances ----------------------------------------------------------------------------------

instance (Spanned a, Eq b) => Eq (Rel a b) where
    u == v = all (\a -> (u $$ a) == (v $$ a)) s_span

instance (Spanned a, Show a, Show b) => Show (Rel a b)  where
    showsPrec _  u s' = show (relToList u) ++ s'

instance (Spanned a, Read a, Read b) => Read (Rel a b)  where
    readsPrec n s = [(listToRel z,s')] where  ((z,s'):_) = readsPrec n s

instance (Spanned a, Spanned b) => Spanned (Rel a b) where
      s_span = map listToRel s_span

instance (Spanned a, Spanned b) => Cpo (Rel a b) where
    top     = Rel (\a -> s_span)
    bot     = Rel (\a -> [])
    range   = s_span 
    u .<. v = all (\a -> (u $$ a) `inc` (v $$ a)) s_span where inc x y = all (`elem` y) x
    up   u  = concat (map up_ s_span) where
	up_  a   = map (set_ a)  s_span
        set_ a b = let v = u $$ a in Rel (\x -> if (x /= a) || (b `elem` v) then v else (b:v))



---------------------------------------------------------------------------------------------------
--- Bags
---------------------------------------------------------------------------------------------------

data Spanned a => Bag a = Bag { bags :: a -> Int }

bagToList :: Spanned a => Bag a ->  [(a,Int)]
bagToList u = filter ((/=0) . snd) (map (\x -> (x, bags u x)) s_span)

listToBag :: Spanned a => [(a,Int)] -> Bag a
listToBag u = Bag (\a -> maybe 0 id (lookup a u))

------- Instances ----------------------------------------------------------------------------------

instance Spanned a => Cpo (Bag a)  where
    top     = Bag (\_ -> maxBound)
    bot     = Bag (\_ -> 0)
    u .<. v = all (\a -> (bags u a) <= (bags v a)) s_span
    up   u  = map suc s_span where
	          suc p = Bag (\a -> let x = bags u a in if a == p then x+1  else x)


instance Spanned a => Eq (Bag a)  where
    u == v = all (\a -> (bags u a) == (bags v a)) s_span

instance (Spanned a, Show a) => Show (Bag a)  where
    showsPrec _  s s' = show (bagToList s) ++ s'

instance (Spanned a, Read a) => Read (Bag a)  where
    readsPrec n s = [(listToBag z,s')] where ((z,s'):_) = readsPrec n s


-----------------------------------------------------------------------------------------------------
-- Relations and ST []
-----------------------------------------------------------------------------------------------------

----- renaming & selectors -- 

st2rel :: ST [] s s -> Rel s s
st2rel = Rel . on 

rel2st ::  Rel s s -> ST [] s s
rel2st = delta . ($$)

get :: (a -> b) -> ST [] s a -> ST [] s b
get =  fmap


-----------------------------------------------------------------------------------------------------
-- More sets functions
-----------------------------------------------------------------------------------------------------

----- power set  


power :: [a] -> [[a]]
power = foldl f [[]] where f u x = u ++ (map (x:) u)

----- transitive closure of a relation

cover :: (Eq a) => (a -> [a]) -> a -> [a]
cover f = \x -> maybe [] fst (loop iter ([],[x]))
    where  iter (u,[])  = Nothing
	   iter (u,a:v) = let
			  h w y = if a == y || y `elem` u || y `elem` v
				  then w else w ++ [y]
                          in Just (a:u , foldl h v (f a))


---- Funny : monadic union and difference of sets

junion, jdiff  :: (Eq a, MonadPlus m) => [a] -> [a] -> m [a]

---- (u `junion` v) ==  return (u `union` v)  if "u" and "v" are disjoint,
---- (u `junion` v) ==  mzero                 otherwise

junion u v = foldl j (return u) v where
-- j  :: m [a] -> a -> m [a]
-- jj :: m [a] -> a -> m [a]
    j u a = u >>= foldl jj (return [a]) where
      jj w b = guard (b/=a) >> w >>= return . (b:)

---- difference

---- (u `diff` v) ==  return (u \\ v)  if "u" contains "v"
---- (u `diff` v) ==  mzero             otherwise

jdiff u v = foldl j (return u) v where
-- j :: Eq a => m [a] -> a -> m [a]
  j u x = u >>= jj  where
     jj  []    = mzero
     jj  (y:w) = if x == y then return w else jj w  >>= return . (y:)
